/*
 DepthBackground.cpp: Background estimation for the depth maps and 
 color images as well as constructing a confidence map.
*/

#include "DepthBackground.h"

/********************************** Main *************************************/
int main(int argc, char** argv)
{
	int arg = 0, index;
	// image file name, segmentation file name, color image file name
	char *img_file[IMAGE_NUM], *seg_file[IMAGE_NUM], *clr_file[IMAGE_NUM];
	// order: depth maps, segmentation file, color images, depth result,
	// color result, confidence map
	IplImage **img, **seg, **clr, *result, *cresult, *coresult;
	uchar **imgdata, **segdata, **clrdata, *resultdata, *cresultdata, *coresultdata;

	while (++arg < argc) {
		// read in the name of depth maps, segmentation fiels and color images
		// from the command line.
		if (arg <= IMAGE_NUM)
			img_file[arg-1] = argv[arg];
		else if (arg > IMAGE_NUM && arg <= IMAGE_NUM*2)
			seg_file[arg-1-IMAGE_NUM] = argv[arg];
		else
			clr_file[arg-1-IMAGE_NUM*2] = argv[arg];
	}

	// initialize the array which will be used in confidence map.
	for (index = 0; index < 196608; index++)
		variance[0][index] = 0; variance[1][index] = 0; variance[2][index] = 0;

	// allocate memory for all images.
	img = (IplImage **)malloc(IMAGE_NUM * sizeof(IplImage *));
	seg = (IplImage **)malloc(IMAGE_NUM * sizeof(IplImage *));
	clr = (IplImage **)malloc(IMAGE_NUM * sizeof(IplImage *));
	imgdata = (uchar **)malloc(IMAGE_NUM * sizeof(uchar *));
	segdata = (uchar **)malloc(IMAGE_NUM * sizeof(uchar *));
	clrdata = (uchar **)malloc(IMAGE_NUM * sizeof(uchar *));
	result = cvCreateImage(cvSize(512, 384), IPL_DEPTH_8U, 3);
	resultdata = (uchar *)result->imageData;
	cresult = cvCreateImage(cvSize(512, 384), IPL_DEPTH_8U, 3);
	cresultdata = (uchar *)cresult->imageData;
	coresult = cvCreateImage(cvSize(512, 384), IPL_DEPTH_8U, 3);
	coresultdata = (uchar *)coresult->imageData;

	for (index = 0; index < IMAGE_NUM; index++) {
		// read the images
		img[index] = cvLoadImage(img_file[index], 1);
		imgdata[index] = (uchar *)img[index]->imageData;
		seg[index] = cvLoadImage(seg_file[index], 1);
		segdata[index] = (uchar *)seg[index]->imageData;
		clr[index] = cvLoadImage(clr_file[index], 1);
		clrdata[index] = (uchar *)clr[index]->imageData;
	}

	// main function for background estimation
	StitchImage(img, imgdata, seg, segdata, clr, clrdata, result, resultdata, cresult, cresultdata, coresult, coresultdata);

	// free up
	cvReleaseImage(img); cvReleaseImage(seg); cvReleaseImage(clr); cvReleaseImage(&result); cvReleaseImage(&cresult);
	return 0;
}
