#include <stdio.h>
#include <stdlib.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>
#include <math.h>
#include <time.h>

// define the number of input images
#define IMAGE_NUM 89

// the array for confidence map
float variance[3][196608];

// this function determines whether two pixels are the same color or not.
// just like the above function, the value of the three channels may not
// be exactly the same, so we use a threshold.
int SameOrNot(IplImage* img, uchar *imgdata, int Coordx1, int Coordy1, int Coordx2, int Coordy2)
{
	if (abs(imgdata[Coordx1*img->widthStep+Coordy1*img->nChannels] - imgdata[Coordx2*img->widthStep+Coordy2*img->nChannels]) < 5 &&
		abs(imgdata[Coordx1*img->widthStep+Coordy1*img->nChannels+1] - imgdata[Coordx2*img->widthStep+Coordy2*img->nChannels+1]) < 5 &&
		abs(imgdata[Coordx1*img->widthStep+Coordy1*img->nChannels+2] - imgdata[Coordx2*img->widthStep+Coordy2*img->nChannels+2]) < 5)
		return 1;
	else
		return 0;
}

// this function determines whether two pixels are similar to each other or not.
// we use a range to determine it. suppose we have two pixels p1 and p2, then
// if |p1-p2|<C*p1, the two pixels are considered to be similar to each other.
// so the constant C is changed according to the illumination condition.
// since the images are taken from the sameviewpoint, we always set it to be 0.9 and 1.1
int IsSimilar(IplImage* img1, uchar *imgdata1, IplImage* img2, uchar *imgdata2, int Coordx, int Coordy)
{
	if (((imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels] > 0.9*imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels] &&
		imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels] < 1.1*imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels]) ||
		abs(imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels] - imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels]) < 20) &&
		((imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels+1] > 0.9*imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels+1] &&
		imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels+1] < 1.1*imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels+1]) ||
		abs(imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels+1] - imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels+1]) < 20) &&
		((imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels+2] > 0.9*imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels+2] &&
		imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels+2] < 1.1*imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels+2]) ||
		abs(imgdata2[Coordx*img2->widthStep+Coordy*img2->nChannels+2] - imgdata1[Coordx*img1->widthStep+Coordy*img1->nChannels+2]) < 20))
		return 1;
	else
		return 0;
}

// linear blending for depth mpas
// since the images are taken from the same viewpoint, we don't need to compute
// the weight using the perspective transform, we only need to use averaging.
void LinearBlend(IplImage** img, uchar** imgdata, int index1, int index2, int* finalvalue, int imgidx1)
{
	int imgidx, flag, index3;
	int weightsum = 0;

	finalvalue[0] = 0; finalvalue[1] = 0; finalvalue[2] = 0;

	// here we go throught each input image, if the pixel in a certain location of
	// some input image is similar to the candidate pixel, then that pixel will be
	// used to do the linear blending.
	for (imgidx = 0; imgidx < IMAGE_NUM; imgidx++) {
		if (flag = IsSimilar(img[imgidx1], imgdata[imgidx1], img[imgidx], imgdata[imgidx], index1, index2) == 1) {
			weightsum++;
			for (index3 = 0; index3 < img[0]->nChannels; index3++)
				finalvalue[index3] = finalvalue[index3] + imgdata[imgidx][index1*img[0]->widthStep+index2*img[0]->nChannels+index3];
		}
	}
	if (weightsum < 0.0000001)
		for (index3 = 0; index3 < img[0]->nChannels; index3++)
			finalvalue[index3] = 0;
	else
		for (index3 = 0; index3 < img[0]->nChannels; index3++)
			finalvalue[index3] = cvRound(finalvalue[index3]/weightsum);
}

// linear blending for color images as well as compute the array for confidence map.
// we still use averaging for linear blending.
// the confidence map is built in this way: for each candidate pixel, there are several
// similar pixels in the same location selected for linear blending, then for those pixels,
// we compute the variance of them.
void LinearBlend_Clr(IplImage** img, uchar** imgdata, IplImage** clr, uchar** clrdata, int index1, int index2, int* finalvalue, int imgidx1)
{
	int imgidx, flag, index3;
	float avg[3];
	int weightsum = 0;

	finalvalue[0] = 0; finalvalue[1] = 0; finalvalue[2] = 0;
	avg[0] = 0; avg[1] = 0; avg[2] = 0;

	// here we go throught each input image, if the pixel in a certain location of
	// some input image is similar to the candidate pixel, then that pixel will be
	// used to do the linear blending. At the same time, the average is computed.
	for (imgidx = 0; imgidx < IMAGE_NUM; imgidx++) {
		if (flag = IsSimilar(img[imgidx1], imgdata[imgidx1], img[imgidx], imgdata[imgidx], index1, index2) == 1) {
			weightsum++;
			for (index3 = 0; index3 < img[0]->nChannels; index3++) {
				finalvalue[index3] = finalvalue[index3] + clrdata[imgidx][index1*img[0]->widthStep+index2*img[0]->nChannels+index3];
				avg[index3] = avg[index3] + clrdata[imgidx][index1*img[0]->widthStep+index2*img[0]->nChannels+index3];
			}
		}
	}
	for (index3 = 0; index3 < img[0]->nChannels; index3++)
		avg[index3] = avg[index3]/weightsum;

	if (weightsum < 0.0000001)
		for (index3 = 0; index3 < img[0]->nChannels; index3++)
			finalvalue[index3] = 0;
	else
		for (index3 = 0; index3 < img[0]->nChannels; index3++)
			finalvalue[index3] = cvRound(finalvalue[index3]/weightsum);

	// after the average is computed, we compute the variance of those pixels in this location
	for (imgidx = 0; imgidx < IMAGE_NUM; imgidx++)
		if (flag = IsSimilar(img[imgidx1], imgdata[imgidx1], img[imgidx], imgdata[imgidx], index1, index2) == 1)
			for (index3 = 0; index3 < img[0]->nChannels; index3++)
				variance[index3][index1*img[0]->width+index2] += (avg[index3] - clrdata[imgidx][index1*img[0]->widthStep+index2*img[0]->nChannels+index3] * 
					(avg[index3] - clrdata[imgidx][index1*img[0]->widthStep+index2*img[0]->nChannels+index3]));

	for (index3 = 0; index3 < img[0]->nChannels; index3++)
		variance[index3][index1*img[0]->width+index2] = variance[index3][index1*img[0]->width+index2]/weightsum;
}

// the main function to do the background estimation.
void StitchImage(IplImage** img, uchar** imgdata, IplImage** seg, uchar** segdata, IplImage** clr, uchar** clrdata, IplImage* result, uchar* resultdata, 
				 IplImage* cresult, uchar* cresultdata, IplImage* coresult, uchar* coresultdata)
{
	CvMat *Tran;
	int imgidx1, imgidx2, index1, index2, index3, index;
	double **cost, mincost;
	int **l, lastidx, flag, finalc[3];
	float smallest = 999999, biggest = -1;

	// Initialization for DP
	// the DP method used here is similar to the assembly-line scheduling problem
	// for the details of that problem, refer to the chapter 15 of the book
	// "Introduction to the algorithm".
	// the matrix cost stores the cost so far to the current coordinate
	// and the matrix l stores the choice of candidate pixel.
	cost = (double **)malloc(IMAGE_NUM * sizeof(double *));
	l = (int **)malloc(IMAGE_NUM * sizeof(int *));
	for (index1 = 0; index1 < IMAGE_NUM; index1++) {
		cost[index1] = (double *)malloc(result->width * sizeof (double));
		l[index1] = (int *)malloc((result->width-1) * sizeof (int));
	}
	Tran = cvCreateMat(IMAGE_NUM * IMAGE_NUM, result->width-1, CV_64F);

	for (index1 = 0; index1 < result->height; index1++) {
		for (index2 = 1; index2 < result->width; index2++) {
			for (imgidx1 = 0; imgidx1 < IMAGE_NUM; imgidx1++) {
				for (imgidx2 = 0; imgidx2 < IMAGE_NUM; imgidx2++) {
					mincost = 0;
					// the smoothness term is the cost between the current pixel and the result
					// that is one scan line above. the smoothness term will not applied on the
					// first scanline because there is no previous scan line.
					if (index1 > 0)
						for (index3 = 0; index3 < result->nChannels; index3++)
							mincost = mincost+fabs(imgdata[imgidx2][index1*result->widthStep+(index2-1)*result->nChannels+index3] -
							resultdata[(index1-1)*result->widthStep+(index2-1)*result->nChannels+index3]);
					// if the adjacent pixels are in the same segment
					// in this case, we assign 0 penality because they
					// are in the same segment, so we only add the smoothness term to it.
					if (imgidx1 == imgidx2 && 
						(flag = SameOrNot(seg[imgidx2], segdata[imgidx2], index1, index2, index1, index2-1)) == 1)
						cvmSet(Tran, imgidx1*IMAGE_NUM+imgidx2, index2-1, mincost/2);
					// if the adjacent pixels are in different segments.
					// in this case, the cost is the color difference plus the smoothness term.
					else {
						for (index3 = 0; index3 < result->nChannels; index3++)
							mincost = mincost+
							fabs(imgdata[imgidx1][index1*result->widthStep+index2*result->nChannels+index3] -
							imgdata[imgidx2][index1*result->widthStep+(index2-1)*result->nChannels+index3])*0.5;
						cvmSet(Tran, imgidx1*IMAGE_NUM+imgidx2, index2-1, mincost/2);
					}
				}
			}
		}

		// before we do the DP, we first assign the cost to be 0 everywhere.
		for (imgidx1 = 0; imgidx1 < IMAGE_NUM; imgidx1++)
			cost[imgidx1][0] = 0;
		// now we apply DP to get the minimum cost.
		// the code comes from the pesdocode in the book "Introduction to algorithms", chapter 15, the
		// first example: assembly-line scheduling.
		for (index2 = 1; index2 < result->width; index2++) {
			for (imgidx1 = 0; imgidx1 < IMAGE_NUM; imgidx1++) {
				mincost = 10000000;
				for (imgidx2 = 0; imgidx2 < IMAGE_NUM; imgidx2++) {
					cost[imgidx1][index2] = cost[imgidx2][index2-1]+
						Tran->data.db[(imgidx1*IMAGE_NUM+imgidx2)*(result->width-1)+(index2-1)];
					if (cost[imgidx1][index2] < mincost) {
						mincost = cost[imgidx1][index2];
						l[imgidx1][index2-1] = imgidx2;
					}
				}
				cost[imgidx1][index2] = mincost;
			}
		}
		mincost = 10000000;
		for (imgidx1 = 0; imgidx1 < IMAGE_NUM; imgidx1++) {
			if (cost[imgidx1][result->width-1] < mincost) {
				mincost = cost[imgidx1][result->width-1];
				lastidx = imgidx1;
			}
		}
		for (index2 = result->width-2; index2 >= 0; index2--) {
			//here l[] is the candidate pixel that has been selected.
			imgidx1 = l[lastidx][index2];
			// perform linear blending after the candidate pixel is selected
			LinearBlend(img, imgdata, index1, index2, finalc, imgidx1);
			for (index3 = 0; index3 < result->nChannels; index3++)
				resultdata[index1*result->widthStep+index2*result->nChannels+index3] = finalc[index3];
			// linear blending for color images as well
			LinearBlend_Clr(img, imgdata, clr, clrdata, index1, index2, finalc, imgidx1);
			for (index3 = 0; index3 < result->nChannels; index3++)
				cresultdata[index1*result->widthStep+index2*result->nChannels+index3] = finalc[index3];
			lastidx = imgidx1;
		}
	}

	// now we normalize the variance of the whole image so that
	// in the confidence map, the pixel with the highest variance
	// will be white.
	for (index = 0; index < 196608; index++) {
		if ((variance[0][index]+variance[1][index]+variance[2][index])/3 < smallest)
			smallest = (variance[0][index]+variance[1][index]+variance[2][index])/3;
		if ((variance[0][index]+variance[1][index]+variance[2][index])/3 > biggest)
			biggest = (variance[0][index]+variance[1][index]+variance[2][index])/3;
	}

	for (index1 = 0; index1 < result->height; index1++) {
		for (index2 = 0; index2 < result->width; index2++) {
			flag = 0;
			for (index3 = 0; index3 < result->nChannels; index3++)
				flag = flag + variance[index3][index1*result->width+index2];
			for (index3 = 0; index3 < result->nChannels; index3++)
				coresultdata[index1*cresult->widthStep+index2*cresult->nChannels+index3] = cvRound(flag*255/(biggest*3));
		}
	}

	cvSaveImage("result_depth.bmp", result);
	cvSaveImage("result_clr.bmp", cresult);
	cvSaveImage("result_conf.bmp", coresult);
	cvReleaseMat(&Tran);
}
